@@ -1,27 +1,28 @@
 #include <pybind11/pybind11.h>
 #include <pybind11/stl.h>
+#include <pybind11/numpy.h>
 
 #include <kompute/Kompute.hpp>
 
 #include "docstrings.hpp"
 
 namespace py = pybind11;
+//used in Core.hpp
+py::object kp_debug, kp_info, kp_warning, kp_error;
 
 PYBIND11_MODULE(kp, m) {
 
-#if KOMPUTE_ENABLE_SPDLOG
-    spdlog::set_level(
-      static_cast<spdlog::level::level_enum>(SPDLOG_ACTIVE_LEVEL));
-#endif
+    // The logging modules are used in the Kompute.hpp file
+    py::module_ logging  = py::module_::import("logging");
+    py::object kp_logger = logging.attr("getLogger")("kp");
+    kp_debug             = kp_logger.attr("debug");
+    kp_info              = kp_logger.attr("info");
+    kp_warning           = kp_logger.attr("warning");
+    kp_error             = kp_logger.attr("error");
+    logging.attr("basicConfig")();
+
+    py::module_ np = py::module_::import("numpy");
 
-    m.def("log_level", [](uint8_t logLevel) {
-#if KOMPUTE_ENABLE_SPDLOG
-            spdlog::set_level(
-              static_cast<spdlog::level::level_enum>(SPDLOG_LEVEL_INFO));
-#else
-            SPDLOG_WARN("SPDLOG not enabled so log level config function not supported");
-#endif
-        });
 
     py::enum_<kp::Tensor::TensorTypes>(m, "TensorTypes", DOC(kp, Tensor, TensorTypes))
         .value("device", kp::Tensor::TensorTypes::eDevice, "Tensor holding data in GPU memory.")
@@ -29,21 +30,35 @@ PYBIND11_MODULE(kp, m) {
         .value("storage", kp::Tensor::TensorTypes::eStorage, "Tensor with host visible gpu memory.")
         .export_values();
 
+
     py::class_<kp::Tensor, std::shared_ptr<kp::Tensor>>(m, "Tensor", DOC(kp, Tensor))
         .def(py::init(
-            [](const std::vector<float>& data) {
-                return std::unique_ptr<kp::Tensor>(new kp::Tensor(data));
-            }), DOC(kp, Tensor, Tensor, 2))
-        .def(py::init(
-            [](const std::vector<float>& data, kp::Tensor::TensorTypes tensorTypes) {
-                return std::unique_ptr<kp::Tensor>(new kp::Tensor(data, tensorTypes));
-            }), "Initialiser with list of data components and tensor GPU memory type.")
+            [np](const py::array_t<float> data, kp::Tensor::TensorTypes tensor_type) {
+                const py::array_t<float> flatdata = np.attr("ravel")(data);
+                const py::buffer_info info        = flatdata.request();
+                const float* ptr                  = (float*) info.ptr;
+                return std::unique_ptr<kp::Tensor>(
+                    new kp::Tensor(std::vector<float>(ptr, ptr+flatdata.size()), tensor_type)
+                );
+            }), 
+            "Construct Tensor with an array as initial data and an optional kp.TensorType (default:device).",
+            py::arg("data"),
+            py::arg("tensor_type") = kp::Tensor::TensorTypes::eDevice
+            )
         .def("data", &kp::Tensor::data, DOC(kp, Tensor, data))
+        .def("numpy", [](kp::Tensor& self) {
+                return py::array(self.data().size(), self.data().data());
+            }, "Returns stored data as a new numpy array.")
         .def("__getitem__", [](kp::Tensor &self, size_t index) -> float { return self.data()[index]; },
                 "When only an index is necessary")
         .def("__setitem__", [](kp::Tensor &self, size_t index, float value) {
                 self.data()[index] = value; })
-        .def("set_data", &kp::Tensor::setData, "Overrides the data in the local Tensor memory.")
+        .def("set_data", [np](kp::Tensor &self, const py::array_t<float> data){
+                const py::array_t<float> flatdata = np.attr("ravel")(data);
+                const py::buffer_info info        = flatdata.request();
+                const float* ptr                  = (float*) info.ptr;
+                self.setData(std::vector<float>(ptr, ptr+flatdata.size()));
+            }, "Overrides the data in the local Tensor memory.")
         .def("__iter__", [](kp::Tensor &self) {
                 return py::make_iterator(self.data().begin(), self.data().end());
             }, py::keep_alive<0, 1>(), // Required to keep alive iterator while exists
@@ -71,19 +86,24 @@ PYBIND11_MODULE(kp, m) {
         .def("map_data_from_host", &kp::Tensor::mapDataFromHostMemory, "Maps data into GPU memory from tensor local data.")
         .def("map_data_into_host", &kp::Tensor::mapDataIntoHostMemory, "Maps data from GPU memory into tensor local data.");
 
+
     py::class_<kp::Sequence, std::shared_ptr<kp::Sequence>>(m, "Sequence")
         .def("init", &kp::Sequence::init, "Initialises Vulkan resources within sequence using provided device.")
+        
         // record
         .def("begin", &kp::Sequence::begin, "Clears previous commands and starts recording commands in sequence which can be run in batch.")
         .def("end", &kp::Sequence::end, "Stops listening and recording for new commands.")
+        
         // eval
         .def("eval", &kp::Sequence::eval, "Executes the currently recorded commands synchronously by waiting on Vulkan Fence.")
         .def("eval_async", &kp::Sequence::evalAsync, "Executes the currently recorded commands asynchronously.")
         .def("eval_await", &kp::Sequence::evalAwait, "Waits until the execution finishes using Vulkan Fence.")
+        
         // status
         .def("is_running", &kp::Sequence::isRunning, "Checks whether the Sequence operations are currently still executing.")
         .def("is_rec", &kp::Sequence::isRecording, "Checks whether the Sequence is currently in recording mode.")
         .def("is_init", &kp::Sequence::isInit, "Checks if the Sequence has been initialized")
+        
         // record
         .def("record_tensor_create", &kp::Sequence::record<kp::OpTensorCreate>,
             "Records operation to create and initialise tensor GPU memory and buffer")
@@ -95,23 +115,38 @@ PYBIND11_MODULE(kp, m) {
             "Records operation to sync tensor(s) from GPU memory to local memory using staging tensors")
         .def("record_algo_mult", &kp::Sequence::record<kp::OpMult>,
             "Records operation to run multiplication compute shader to two input tensors and an output tensor")
-        .def("record_algo_file", &kp::Sequence::record<kp::OpAlgoBase, std::string>,
-            "Records an operation using a custom shader provided from a shader path")
+        .def("record_algo_file", [](kp::Sequence &self, 
+                                    std::vector<std::shared_ptr<kp::Tensor>> tensors,
+                                    const std::string& file_path,
+                                    std::tuple<uint32_t,uint32_t,uint32_t> work_group) -> bool {
+                const kp::OpAlgoBase::KomputeWorkgroup wgroup{
+                    std::get<0>(work_group), std::get<1>(work_group), std::get<2>(work_group),
+                };
+                return self.record<kp::OpAlgoBase>(tensors, file_path, wgroup);
+            },
+            "Records an operation using a custom shader provided from a shader path",
+            py::arg("tensors"), py::arg("file_path"), py::arg("work_group") = std::make_tuple(0,0,0)  )
         .def("record_algo_data", [](kp::Sequence &self,
                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
-                                    py::bytes &bytes) -> float {
+                                    py::bytes &bytes,
+                                    std::tuple<uint32_t,uint32_t,uint32_t> work_group) -> bool {
                 // Bytes have to be converted into std::vector
                 py::buffer_info info(py::buffer(bytes).request());
                 const char *data = reinterpret_cast<const char *>(info.ptr);
                 size_t length = static_cast<size_t>(info.size);
+                const kp::OpAlgoBase::KomputeWorkgroup wgroup{
+                    std::get<0>(work_group), std::get<1>(work_group), std::get<2>(work_group),
+                };
                 return self.record<kp::OpAlgoBase>(
-                    tensors,
-                    std::vector<char>(data, data + length));
+                    tensors, std::vector<char>(data, data + length), wgroup
+                );
             },
-            "Records an operation using a custom shader provided as raw string or spirv bytes")
+            "Records an operation using a custom shader provided as spirv bytes",
+            py::arg("tensors"), py::arg("bytes"), py::arg("work_group") = std::make_tuple(0,0,0)  )
         .def("record_algo_lro", &kp::Sequence::record<kp::OpAlgoLhsRhsOut>,
             "Records operation to run left right out operation with custom shader");
 
+
     py::class_<kp::Manager>(m, "Manager")
         .def(py::init(), "Default initializer uses device 0 and first compute compatible GPU queueFamily")
         .def(py::init(
@@ -128,12 +163,14 @@ PYBIND11_MODULE(kp, m) {
         .def("build_tensor", &kp::Manager::buildTensor, 
                 py::arg("data"), py::arg("tensorType") = kp::Tensor::TensorTypes::eDevice,
                 "Build and initialise tensor")
+        
         // Await functions
         .def("eval_await", &kp::Manager::evalOpAwait,
                 py::arg("sequenceName"), py::arg("waitFor") = UINT64_MAX,
                 "Awaits for asynchronous operation on a named Sequence")
         .def("eval_await_def", &kp::Manager::evalOpAwaitDefault,
                 py::arg("waitFor") = UINT64_MAX, "Awaits for asynchronous operation on the last anonymous Sequence created")
+        
         // eval default
         .def("eval_tensor_create_def", &kp::Manager::evalOpDefault<kp::OpTensorCreate>,
             "Evaluates operation to create and initialise tensor GPU memory and buffer with new anonymous Sequence")
@@ -149,6 +186,13 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates an operation using a custom shader provided from a shader path with new anonymous Sequence")
         .def("eval_algo_str_def", &kp::Manager::evalOpDefault<kp::OpAlgoBase, std::vector<char>>,
             "Evaluates an operation using a custom shader provided as string provided as list of characters with new anonymous Sequence")
+        .def("eval_algo_str_def", [](kp::Manager &self,
+                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
+                                     const std::string& shader_str){
+                const std::vector<char> shader_vec(shader_str.begin(), shader_str.end());
+                self.evalOpDefault<kp::OpAlgoBase>(tensors, shader_vec);
+        },
+            "Evaluates an operation using a custom shader provided as string with a new anonymous Sequence")
         .def("eval_algo_data_def", [](kp::Manager &self,
                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
                                     py::bytes &bytes) {
@@ -163,6 +207,7 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates an operation using a custom shader provided as spirv bytes with new anonymous Sequence")
         .def("eval_algo_lro_def", &kp::Manager::evalOpDefault<kp::OpAlgoLhsRhsOut>,
             "Evaluates operation to run left right out operation with custom shader with new anonymous Sequence")
+        
         // eval
         .def("eval_tensor_create", &kp::Manager::evalOp<kp::OpTensorCreate>,
             "Evaluates operation to create and initialise tensor GPU memory and buffer with explicitly named Sequence")
@@ -178,6 +223,14 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates an operation using a custom shader provided from a shader path with explicitly named Sequence")
         .def("eval_algo_str", &kp::Manager::evalOp<kp::OpAlgoBase, std::vector<char>>,
             "Evaluates an operation using a custom shader provided as string provided as list of characters with explicitly named Sequence")
+        .def("eval_algo_str", [](kp::Manager &self,
+                                    std::vector<std::shared_ptr<kp::Tensor>> tensors,
+                                    const std::string& sequenceName,
+                                    const std::string& shader_str) {
+                const std::vector<char> shader_vec(shader_str.begin(), shader_str.end());
+                self.evalOp<kp::OpAlgoBase>(tensors, sequenceName, shader_vec);
+            },
+            "Evaluates an operation using a custom shader provided as string with explicitly named Sequence")
         .def("eval_algo_data", [](kp::Manager &self,
                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
                                     std::string sequenceName,
@@ -194,6 +247,7 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates an operation using a custom shader provided as spirv bytes with explicitly named Sequence")
         .def("eval_algo_lro", &kp::Manager::evalOp<kp::OpAlgoLhsRhsOut>,
             "Evaluates operation to run left right out operation with custom shader with explicitly named Sequence")
+        
         // eval async default
         .def("eval_async_tensor_create_def", &kp::Manager::evalOpAsyncDefault<kp::OpTensorCreate>,
             "Evaluates asynchronously operation to create and initialise tensor GPU memory and buffer with anonymous Sequence")
@@ -209,6 +263,13 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates asynchronously an operation using a custom shader provided from a shader path with anonymous Sequence")
         .def("eval_async_algo_str_def", &kp::Manager::evalOpAsyncDefault<kp::OpAlgoBase, std::vector<char>>,
             "Evaluates Asynchronously an operation using a custom shader provided as string provided as list of characters with new anonymous Sequence")
+        .def("eval_async_algo_str_def", [](kp::Manager &self,
+                                           std::vector<std::shared_ptr<kp::Tensor>> tensors,
+                                           const std::string& shader_str) {
+                const std::vector<char> shader_vec(shader_str.begin(), shader_str.end());
+                self.evalOpAsyncDefault<kp::OpAlgoBase>(tensors, shader_vec);
+            },
+            "Evaluates Asynchronously an operation using a custom shader provided as string with new anonymous Sequence")
         .def("eval_async_algo_data_def", [](kp::Manager &self,
                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
                                     py::bytes &bytes) {
@@ -223,6 +284,7 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates asynchronously an operation using a custom shader provided as raw string or spirv bytes with anonymous Sequence")
         .def("eval_async_algo_lro_def", &kp::Manager::evalOpAsyncDefault<kp::OpAlgoLhsRhsOut>,
             "Evaluates asynchronously operation to run left right out operation with custom shader with anonymous Sequence")
+        
         // eval async
         .def("eval_async_tensor_create", &kp::Manager::evalOpAsync<kp::OpTensorCreate>,
             "Evaluates asynchronously operation to create and initialise tensor GPU memory and buffer with explicitly named Sequence")
@@ -238,6 +300,14 @@ PYBIND11_MODULE(kp, m) {
             "Evaluates asynchronously an operation using a custom shader provided from a shader path with explicitly named Sequence")
         .def("eval_async_algo_str", &kp::Manager::evalOpAsync<kp::OpAlgoBase, std::vector<char>>,
             "Evaluates Asynchronous an operation using a custom shader provided as string provided as list of characters with explicitly named Sequence")
+        .def("eval_async_algo_str", [](kp::Manager &self,
+                                    std::vector<std::shared_ptr<kp::Tensor>> tensors,
+                                    const std::string& sequenceName,
+                                    const std::string& shader_str) {
+                const std::vector<char> shader_vec(shader_str.begin(), shader_str.end());
+                self.evalOpAsync<kp::OpAlgoBase>(tensors, sequenceName, shader_vec);
+            },
+            "Evaluates Asynchronous an operation using a custom shader provided as string with explicitly named Sequence")
         .def("eval_async_algo_data", [](kp::Manager &self,
                                     std::vector<std::shared_ptr<kp::Tensor>> tensors,
                                     std::string sequenceName,
